/* Выборки из эпизодической памяти Моделей понимания
ВЫБОРКИ КАДРОВ С ЭКСТРЕМАЛЬНЫМИ ОБЪЕКТАМИ

Главное отличие от объектов внимания в кадрах с правилами: тут могут быть не полные объекты, а их части,
что может давать важную инфу.

Пока не очень понятно, как это применять.
Разве что посмотреть, что делалось каждый раз после фиксации такого-то объекта внимания,
т.е. просканировать кадры после таких фиксаций.
Ну и для категоризации объектов для разных условий:
какую они значимость демонстрируют.

*/

package psychic

import "BOT/lib"

//////////////////////////////////////////////////////////////

/*
	Выборка эпизодов Модели понимания экстремально значащего образа objID типа kind

Найти все эпизоды с экстремальными объектами внимания,
т.е. с Trigger == objID и node.Action == kind (тип объекта внимания)
Ищем только объекты Модели понимания c node.PARAMS[0][2]>0

Фактически означает пройти все дерево эпиз.памяти... Пока нигде не применяется и не нужно.
*/
func getUnderstandingModel(objID int, kind int) {
	UnderstandingModel = nil

	getObjModel(&EpisodicTree, objID, kind)
}
func getObjModel(node *EpisodicTreeNode, objID int, kind int) {
	// вытащить все ID узлов, в которых есть условие, начиная с node
	//func getIdArr(node *EpisodicTreeNode,Trigger int,Action int,oldOnly bool){
	for _, child := range node.Children {
		if len(child.PARAMS) > 0 { // узел с прописанным PARAMS

			if child.PARAMS[2] != 0 { // это не объект
				continue
			}

			if node.Trigger != objID || (node.Action) != kind {
				continue
			}

			//ePsrsIdArr=append(ePsrsIdArr,child.ID)
			om := ObjModel{child.ID, child.Trigger, child.Action, child.PARAMS[0]}
			UnderstandingModel = append(UnderstandingModel, om)
		}
		// продолжим рекурсивно искать в нем далее
		getObjModel(&child, objID, kind)
	}
	return
}

/////////////////////////////////////////////////////////////////////////////////

/*
	Выборка эпизодов Модели понимания экстремально значащего образа objID типа kind

ТОЧНО ДЛЯ ТЕКУЩИХ УСЛОВИЙ NodeAID,NodeSID,NodePID
Найти все эпизоды с экстремальными объектами внимания,
т.е. с Trigger == objID и node.Action == kind (тип объекта внимания)
Ищем только объекты Модели понимания c node.PARAMS[0][2]>0

Очень быстрая.
Реализация по аналогии с func getEpisodesArrFromConditions(level int,Trigger int,Action int)
так что если нужно не столь точное сочетание условий, можно использовать пример оттуда.
*/
func getUnderstandingModelExactly(objID int, kind int) {
	if objID == 0 {
		lib.TodoPanic("НУЛЕВОЙ ПАРАМЕТР objID в func getUnderstandingModelExactly!")
		return
	}
	UnderstandingModel = nil

	cond := []int{detectedActiveLastNodID,
		detectedActiveLastUnderstandingNodID,
		detectedActiveLastProblemNodID}
	id, _ := findEpisodicBrange(0, cond, &EpisodicTree)
	if id == 0 {
		return
	}
	node, ok := ReadeEpisodicTreeNodeFromID(id)
	if !ok {
		return
	}

	scanChildrens(node)
}

// вытащить все ID узлов, в которых есть условие, начиная с node
func scanChildrens(node *EpisodicTreeNode) {
	for _, child := range node.Children {

		if len(child.PARAMS) > 0 { // узел с прописанным PARAMS
			if child.PARAMS[2] == 0 {
				continue
			}
			om := ObjModel{child.ID, child.Trigger, child.Action, child.PARAMS[0]}
			UnderstandingModel = append(UnderstandingModel, om)
		}
		// продолжим рекурсивно искать в нем далее
		scanChildrens(&child)
	}
	return
}

/////////////////////////////////////////////////////////////////////////////////

/*
	Выборка эпизодов Модели понимания экстремально значащего образа objID типа kind

ПРИБЛИЗИТЕЛЬНО ДЛЯ ТЕКУЩИХ УСЛОВИЙ NodeAID
Найти все эпизоды с экстремальными объектами внимания,
т.е. с Trigger == objID и node.Action == kind (тип объекта внимания)
Ищем только объекты Модели понимания c node.PARAMS[0][2]>0

Очень быстрая.
Реализация по аналогии с func getEpisodesArrFromConditions(level int,Trigger int,Action int)
так что если нужно не столь точное сочетание условий, можно использовать пример оттуда.
*/
func getUnderstandingModelApproximately(objID int, kind int) {
	if objID == 0 {
		lib.TodoPanic("НУЛЕВОЙ ПАРАМЕТР objID в func getUnderstandingModelApproximately!")
		return
	}
	UnderstandingModel = nil

	cond := []int{detectedActiveLastNodID}
	id, _ := findEpisodicBrange(0, cond, &EpisodicTree)
	if id == 0 {
		return
	}
	node, ok := ReadeEpisodicTreeNodeFromID(id)
	if !ok {
		return
	}

	scanChildrens(node)
}

/////////////////////////////////////////////////////////////////////////////////

/*
	Выборка ВСЕХ эпизодов Модели понимания экстремально значащих образов objID типа kind

Фактически означает пройти все дерево эпиз.памяти...Применяется в Пульте на стр. Модели понимания.
В UnderstandingModelArr [][]int индекс - ID ЭкстремальноЗначащих Объектов, значение - массив ID кадров эпиз.памяти
*/
func getAllExtrОbjIDarr() {
	UnderstandingModelArr = nil

	getAlloIDarr(&EpisodicTree)
}
func getAlloIDarr(node *EpisodicTreeNode) {
	for _, child := range node.Children {
		if len(child.PARAMS) > 0 { // узел с прописанным PARAMS
			if child.PARAMS[2] == 0 {
				continue
			}
			addUnderstandingModelArr(child.Trigger)
			UnderstandingModelArr[child.Trigger] = append(UnderstandingModelArr[child.Trigger], child.ID)
		}
		// продолжим рекурсивно искать в нем далее
		getAlloIDarr(&child)
	}
	return
}

/////////////////////////////////////////////////////////////
